import sys
from typing import Type

from anyio import create_task_group
from fps import Module

from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.terminals import Terminals, TerminalServer

from .routes import _Terminals

_TerminalServer: Type[TerminalServer]
if sys.platform == "win32":
    from .win_server import _TerminalServer
else:
    from .server import _TerminalServer


class TerminalsModule(Module):
    async def prepare(self) -> None:
        app = await self.get(App)
        auth = await self.get(Auth)  # type: ignore[type-abstract]

        self.terminals = _Terminals(app, auth, _TerminalServer)
        self.put(self.terminals, Terminals)
        async with create_task_group() as tg:
            tg.start_soon(self.terminals.start)
            self.done()

    async def stop(self) -> None:
        await self.terminals.stop()



import os
from functools import partial

from anyio import create_task_group, to_thread
from fastapi import WebSocketDisconnect
from winpty import PTY  # type: ignore

from jupyverse_api.terminals import TerminalServer


def open_terminal(command="C:\\Windows\\System32\\cmd.exe", columns=80, lines=24):
    env = "\0".join([f"{k}={v}" for k, v in os.environ.items()]) + "\0"
    process = PTY(columns, lines)
    process.spawn(command, env=env)
    return process


class _TerminalServer(TerminalServer):
    def __init__(self):
        self.process = open_terminal()
        self.websockets = []

    async def serve(self, websocket, permissions) -> None:
        self.websocket = websocket
        self.permissions = permissions
        self.websockets.append(websocket)

        await websocket.send_json(["setup", {}])

        async with create_task_group() as tg:
            self.task_group = tg
            tg.start_soon(self.send_data)
            tg.start_soon(self.recv_data)

    async def stop(self) -> None:
        self.task_group.cancel_scope.cancel()

    async def send_data(self):
        while True:
            try:
                data = await to_thread.run_sync(partial(self.process.read, blocking=True))
            except Exception:
                await self.websocket.send_json(["disconnect", 1])
                return
            else:
                for websocket in self.websockets:
                    await websocket.send_json(["stdout", data])

    async def recv_data(self):
        can_execute = self.permissions is None or "execute" in self.permissions.get("terminals", [])
        try:
            while True:
                msg = await self.websocket.receive_json()
                if can_execute:
                    if msg[0] == "stdin":
                        self.process.write(msg[1])
                    elif msg[0] == "set_size":
                        self.process.set_size(msg[2], msg[1])
        except WebSocketDisconnect:
            self.quit(self.websocket)
            self.task_group.cancel_scope.cancel()

    def quit(self, websocket):
        self.websockets.remove(websocket)
        if not self.websockets:
            self.task_group.cancel_scope.cancel()
            del self.process



from datetime import datetime, timezone
from http import HTTPStatus
from typing import Any, Dict, Type

import structlog
from anyio import Event, create_task_group
from fastapi import Response

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.terminals import Terminal, Terminals, TerminalServer

TERMINALS: Dict[str, Dict[str, Any]] = {}

log = structlog.get_logger()


class _Terminals(Terminals):
    def __init__(self, app: App, auth: Auth, _TerminalServer: Type[TerminalServer]) -> None:
        super().__init__(app=app, auth=auth)
        self.TerminalServer = _TerminalServer
        self.stop_event = Event()

    async def start(self):
        await self.stop_event.wait()

    async def stop(self):
        async with create_task_group() as tg:
            for terminal in TERMINALS.values():
                tg.start_soon(terminal["server"].stop)
        self.stop_event.set()

    async def get_terminals(
        self,
        user: User,
    ):
        return [terminal["info"] for terminal in TERMINALS.values()]

    async def create_terminal(
        self,
        user: User,
    ):
        name_int = 1
        while True:
            if str(name_int) not in TERMINALS:
                break
            name_int += 1
        name = str(name_int)
        log.info("Creating terminal", name=name)
        terminal = Terminal(
            name=name,
            last_activity=datetime.now(tz=timezone.utc).isoformat().replace("+00:00", "Z"),
        )
        server = self.TerminalServer()
        TERMINALS[name] = {"info": terminal, "server": server}
        log.info("Terminal created", name=name)
        return terminal

    async def delete_terminal(
        self,
        name: str,
        user: User,
    ):
        log.info("Deleting terminal", name=name)
        if name in TERMINALS:
            for websocket in TERMINALS[name]["server"].websockets:
                TERMINALS[name]["server"].quit(websocket)
            del TERMINALS[name]
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def terminal_websocket(
        self,
        name,
        websocket_permissions,
    ):
        if websocket_permissions is None:
            return
        websocket, permissions = websocket_permissions
        await websocket.accept()
        if name not in TERMINALS:
            return

        await TERMINALS[name]["server"].serve(websocket, permissions)
        if name in TERMINALS:
            TERMINALS[name]["server"].quit(websocket)
            if not TERMINALS[name]["server"].websockets:
                del TERMINALS[name]
        log.info("Terminal exited", name=name)
